#!/bin/tclsh
#
#  Sitestat Invite script (rev 1)
#
#  setup: invic is the array which will store the particular invite
#    commands with all the other info necessary for sitestat to be
#    able to invite, the format is simply:
#
#      set invic(destination) "botname #channel"
#
#    ...the destination can be a simple word or combination of comma
#    seperated ones such like areaname,network so an example might
#    look like one of these:
#
#      set invic(mp3,efnet) "efnetbot #blah"
#      set invic(mp3) "dalbot #music4us"
#      set invic(staff) "efnetbot #private.site.chan"
#
#  note: ideally you'll want short simple and logical destination names,
#    users are not generally fond of excessive typing - and have memory
#    spans of minutes so will likely forget complex ones. Also the first
#    defined destination will be the default if none is specified.
#
#  special note: there is a special destination "staff" which will be 
#    treated as a private channel, only those users in group STAFF will
#    be able to invoke this invite.
#

set invic(home) "IMPERIUM #imperium"
set invic(itr) "IMPERIUM #itr"

set denyflag "8" ;# deny anyone with this flag from inviting

set mlog "/ftp-data/logs/glftpd.log"  ;# location of the log

proc invus {{target ""} {dest ""}} {
	global invic mlog env denyflag
	set user $env(USER)
	set group $env(GROUP)
	set flags $env(FLAGS)
	set dlist [array names invic] ; set ddest [lindex $dlist 0]
	puts "-\[Sitestat-invite\]-----------------------"
	if {$target == "" || $target == "help"} {
		puts "Syntax: site invite <nick> <destination>"
		puts ""
		puts "valid dest's: $dlist"
		puts "default dest: $ddest"
	} elseif [regexp $denyflag $flags] {
		puts "This account isn't allowed to site invite"
		puts ""
	} else {
		if {$dest == ""} {set dest $ddest}
		set pos [lsearch $dlist $dest]
		if {$pos == -1} {
			puts "Invalid choice of vacation spot..."
			puts ""
			puts "valid dest's: $dlist"
			puts "default dest: $ddest"
		} elseif {$dest == "staff" && $group != "STAFF"} {
			puts "Only STAFF can invite to private channel..."
			puts ""
			puts "valid dest's: $dlist"
			puts "default dest: $ddest"
		} else {
			foreach {xbot chan} $invic($dest) {}
			set timestamp [clock format [clock seconds] -format "%a %b %d %T %Y"]
			puts "Inviting $target... via $xbot"
			set of [open $mlog a]
			puts $of "$timestamp INVITE: \{$xbot\} \{$target\} \{$user\} \{$group\} \{$chan\}"
			close $of
		}
	}
}

set args [lindex [list $argv] 0]
set target [lindex $args 0]
set dest [string tolower [lindex $args 1]]
invus $target $dest
