#---------------------------------------------------------------------------------------
#  Name: web.tcl for movie2.tcl
#  Version: 1.0.6m
#  Author: GOD-EMPEROR
#---------------------------------------------------------------------------------------
#  Description:
#    web.tcl is a light and simple, single connection text http client library for use
#    with other controlling scripts, compatible with eggdrop ... blah blah ...
#
#  Features:
#    multiple proxies, will pick one randomly 'set uproxy proxy.boo.com:8080 ...'
#    non-blocking operation, default 25sec timeout on connection failure
#    authorized login, with base64 encoding username and password
#    default port for host if none given, will not add port to referer if its 80
#    fully supports http 1.1 chunking, and http 1.0 explicit sizing
#    debug option for eggdrops to view all http transactions
#    custom definable agent, and http version string
#    supports both binary and ascii data
#    supports any script/cgi arguments to pass appropriately for POST operations
#    small footprint, and lastly: FAAAAAAST
#
#  License: 
#    go read the GPL... use this script at your own risk, ie. no warranty or liability
#    feel free to use and modify this code to integrate, as long as credit is given...
#
#  Acknowledgements:
#    a wee portion of base64: Stephen Uhler / Brent Welch (c) 1997 Sun Microsystems
#---------------------------------------------------------------------------------------
#  Usage: 
#    web::htopen <method> <host>:<port> <document> <cgi values> <referer> <user>:<pass>
#  Returns: 
#    on successful download, the document contents
#    on failure the returned string will be the cgi values if given, blank otherwise
#  Examples:
#    set webpage [web::htopen GET www.cnn.com:80 /]
#    set data [web::htopen POST myweb.dyndns.org /cgi loc=home&id=146511 {} lewser:phat]
#---------------------------------------------------------------------------------------
#  Changelog:
#    1.06  adjustments galore, including oddball imdb chunking support
#    1.05  added support for resuming
#    1.04  added 'adaptive timeout' option
#    1.03  fixed an issue with binary data being truncated
#    1.02  added 'improper chunking' check
#    1.01  improved chunking code
#    1.00  initial release

namespace eval web {
#---------------------------------------------------------------------------------------
#  Here's some internal values that can be overridden by calling scripts as needed
#
#  note: to change these externally, add web:: to each var, eg. 'set web::uproxy ...'
#
set uproxy ""
set timeout 25
set adaptive 1
set debug 0
set htbuffer(agent) "web.tcl/1.0 (compatible; UNIX; U)  \[en\]"
set htbuffer(proto) "HTTP/1.1"
set htbuffer(resume) 0
set htbuffer(chunking) 1 
#
#
################################################################# begin code.

namespace export htopen
package provide web.tcl 1.0
set ver "web.tcl/1.0 (p.4) library by GOD-EMPEROR"
proc enc64 {text {chars "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"}} {
  set encoded "" ; set y 0
  for {set i 0} {$i < [string length $text]} {incr i} {
	binary scan [string index $text $i] c x
	if {$x < 0} {set x [expr $x + 256]}
	set y [expr ($y << 8) + $x]
	if {[expr $i % 3] == 2}  {
append encoded [string index $chars [expr ($y & 0xfc0000) >> 18]] [string index $chars [expr ($y & 0x03f000) >> 12]] [string index $chars [expr ($y & 0x000fc0 ) >> 6]] [string index $chars [expr ($y & 0x00003f)]]
	set y 0
	}
  }
  if {[expr $i % 3] == 1} {
	set y [expr $y << 4]
append encoded [string index $chars [expr ($y & 0x000fc0) >> 6]] [string index $chars [expr ($y & 0x00003f)]] "=="
  }
  if {[expr $i % 3] == 2} {
	set y [expr $y << 2] 
append encoded [string index $chars [expr ($y & 0x03f000) >> 12]] [string index $chars [expr ($y & 0x000fc0) >> 6]] [string index $chars [expr ($y & 0x00003f)]] =
  }
  return $encoded
}


proc webputs {iport string} {
  variable debug
  if {$debug} {putlog "*request* $string"}
  puts $iport $string
}

proc htopen {method url {page "/"} {args ""} {referer ""} {auth ""}} {
  variable istate
  variable idata
  variable uproxy
  variable htbuffer
  variable iport
  variable iware
  variable iwait
  variable timeout
  variable debug
  set idata $args
  set pord 80
  set url [split $url ":"]
  if {[llength $url] > 1} {
    set host [lindex $url 0] ; set pord [lindex $url 1]
  } else {
    set host $url ; set pord 80
  }
			      
  set iware $method
  set htbuffer(size) 0
  set htbuffer(buffer) 0
  set htbuffer(page) ""
  set htbuffer(expire) [clock seconds]
  set htbuffer(chunk) 8192
  set htbuffer(referer) $referer
  if {$auth != ""} {set htbuffer(auth) [enc64 $auth]} else {set htbuffer(auth) ""}
  if {$referer != ""} {set htbuffer(referer) $referer}
  if {$uproxy == ""} {
    set htbuffer(url) $host
    append htbuffer(page) $page
## direct
    if {$debug} {putlog "web: connecting to $host $pord ..."}
    set rc [catch {socket -async $host $pord} iport]

  } else {
    set htbuffer(url) "$host"
    append htbuffer(page) "http://$host"
    if {[string index $page 0] != "/"} {append htbuffer(page) "/"}
    append htbuffer(page) $page
## pick a proxy
    set num [expr round([expr rand() * [llength $uproxy]])] ; if {$num != 0} {incr num -1}
    foreach {host pord} [split [lindex $uproxy $num] ":"] {}
    if {$debug} {putlog "web: connecting to $host $pord ..."}
    set rc [catch {socket -async $host $pord} iport]
  }
  if {$pord != 80} {append htbuffer(url) ":$pord"}
  if {$rc != 1} {
	if {$debug} {putlog "web: opening connection $iport"}
	fconfigure $iport -blocking 0 -buffering line -translation {auto crlf}

	fileevent $iport writable web::alive
 	set iwait [after [expr $timeout * 1000] {set iware "timeout" ; set istate 1}]
	if {$debug} {putlog "web: spawned $iwait"}

	vwait istate
	if {$debug} {putlog "web: wait done >> $iware ([string length $idata])"}
	iclose

	if {$debug} {putlog "web: completed ([expr [clock seconds] - $htbuffer(expire)]s)"}
	return $idata
  } else {
	iclose
	if {$debug} {putlog "web: $iport error ([expr [clock seconds] - $htbuffer(expire)]s)"}
  }
  return ""
}

proc alive {} {
  variable istate
  variable iport
  variable htbuffer
  variable debug
  if {[eof $iport] || [fconfigure $iport -error] != ""} {
    if {$debug} {putlog "web: [fconfigure $iport -error] error ([expr [clock seconds] - $htbuffer(expire)]s)"}
  } else {
    fileevent $iport writable {}
    htget
  }
}

proc iclose {} {
  variable iport
  fileevent $iport writable {}
  fileevent $iport readable {}
  variable debug
  variable iwait
  variable istate
  variable debug
  variable htbuffer 
  set htbuffer(resume) 0
  catch {after cancel [set iwait]}
  if {$debug} {putlog "web: closing connection"}
  catch {close $iport} msg
  if [info exists istate] {unset istate}
}

proc htget {} {
  variable iware
  variable rtype
  variable idata
  variable iport
  variable htbuffer
  variable istate
  variable debug
  variable timeout
  variable iwait
  if {[lsearch [file channels] $iport] == -1 || [eof $iport]} {set iware "closed" ; uplevel #0 {set istate 1}}
  switch -- $iware {
	POST -
	GET {
		variable uproxy
		if {$debug} {putlog "web: sending request header"}
		webputs $iport "$iware $htbuffer(page) $htbuffer(proto)"
		webputs $iport "User-Agent: $htbuffer(agent)"
		webputs $iport "Host: $htbuffer(url)"
		webputs $iport "Accept: text/html, */*"
		webputs $iport "Accept-Language: en"
		webputs $iport "Accept-Encoding: identity, *;q=0"
		webputs $iport "Accept-Charset: windows-1252;q=1.0, iso-8859-1;q=0.6, *;q=0.1"
		if {$htbuffer(referer) != ""} {webputs $iport "Referer: $htbuffer(referer)"}
		if {$uproxy != ""} {
			webputs $iport "Cache-Control: no-cache"
		}
		if {$htbuffer(resume) != 0} {
			webputs $iport "Range: bytes=$htbuffer(resume)-"
		}
		
		if {$htbuffer(chunking) == 1} {
			webputs $iport "Connection: Keep-Alive, TE"
			webputs $iport "TE: chunked, identity"
		} else {
			webputs $iport "Connection: Keep-Alive, TE"
			webputs $iport "TE:"
		}
		set args [string length $idata]
		if {$args != 0} {
			webputs $iport "Content-type: application/x-www-form-urlencoded"
			webputs $iport "Content-length: $args"
		}
		if {$htbuffer(auth) != ""} {webputs $iport "Authorization: Basic $htbuffer(auth)"}
		webputs $iport ""
		flush $iport

		set rtype "header"
		set iware "read_text"
		fileevent $iport readable web::htget

		set htbuffer(chunking) 0
	}
	body {
		webputs $iport $idata
		flush $iport
		set iware "read_text"
		set rtype "reply"
		set idata ""
		fileevent $iport writable {}
		fileevent $iport readable web::htget
	}
	read_text {
		set bufline ""
		if [catch {gets $iport bufline} number] {set iware "closed" ; uplevel #0 {set istate 1} ; return}  ;################
		if [fblocked $iport] {return}
		switch $rtype {
			reply {
				if {$debug} {putlog "*$rtype* $bufline"}
				switch -exact [lindex $bufline 0] {
					Content-Length: {
						set size [lindex $bufline 1]
						set htbuffer(size) $size
						set htbuffer(chunk) $size
						set htbuffer(chunking) -3
					}
					Accept-Ranges: -
					Content-Range: {
						if {[string equal -nocase "bytes" [lindex $bufline 1]]} {set htbuffer(resume) 0}
					}
					Connection: { 
						if {[lindex $bufline 1] == "close"} {incr htbuffer(chunking)}
					}
					Transfer-Encoding: {
						if {[lindex $bufline 1] == "chunked"} {incr htbuffer(chunking)}
					}
					{} {
					#####if {$htbuffer(resume) != 0} {set idata "" ; uplevel #0 {set istate 1} ; return}
						if {$htbuffer(chunking) > 0} {
							set rtype "chunksize"
						} else {
							set iware "read_bin"
							set rtype "done"
						}
						if {$htbuffer(code) == 302} {set iware "redirect ignored" ; uplevel #0 {set istate 1} ; return}
						fconfigure $iport -translation binary
					}
				}
			}
			header {
				if {$bufline == ""} {return}
				foreach {vers rc} $bufline {break}
				fileevent $iport readable {}
				if ![string match "HTTP/1.?" $vers] {
					if {$debug} {putlog "web: bad or unknown response - $bufline"}
					set rc 500
				} else {
					if {$debug} {putlog "*server* $bufline"}
				}
				switch -- $rc { 
					100 -
					206 -
					302 -
					200 {
						set iware "body"
						fileevent $iport readable {}
						fileevent $iport writable web::htget
					}
					default {
						set iware "server error"
						uplevel #0 {set istate 1}
					}
				}
				set htbuffer(code) $rc
			}
			chunksize {
				if [regexp {[A-Fa-f0-9]+} $bufline chunkline] {
					##if ![string equal $bufline $chunkline] {puts "chunking mismatch" ; return}
					set chunk [expr 1 * (0x$chunkline + 1)]
					fconfigure $iport -translation binary
					set iware "read_bin"
					set htbuffer(chunk) $chunk
				} else {
					return ;#########set chunk -1 
				}
				if {$chunk == 0} {set iware "complete" ; uplevel #0 {set istate 1} ; return} ;#############
				if {$debug} {putlog "web: expecting chunk --> $chunk"}
			}
		}
	}
	read_bin {
		variable adaptive
		set bufline "" ; set chunk $htbuffer(chunk)
		set bufline [read $iport $chunk]
		append idata $bufline
		set rs [string length $bufline]
		incr htbuffer(buffer) $rs
		if {$rs < $chunk} {
			incr chunk [expr $rs * -1]
			if {$debug} {putlog "*$rtype* partial chunk ($rs) received: $htbuffer(buffer)"}
			set htbuffer(chunk) $chunk
			if {$adaptive} {
				catch {after cancel [set iwait]}
				set iwait [after [expr $timeout * 1000] {set istate 1}]
				##if {$debug} {putlog "web: timeout reset to [set timeout]s"}
			}
		} else {
			if {$debug} {putlog "*$rtype* complete chunk ($rs) received: $htbuffer(buffer)"}
			if {$htbuffer(chunking) > 0} {
				fconfigure $iport -translation {auto}
				set iware "read_text"
				set rtype "chunksize"
			} else {
				set iware "complete"
				uplevel #0 {set istate 1} ; return  ;#############
			}
		}
	}
  }
}
################### end web namespace ####################
}



